@@ -410,6 +410,10 @@ class PanClient(AsyncClient):
             except (asyncio.CancelledError, KeyboardInterrupt):
                 return
 
+    @property
+    def has_been_synced(self) -> bool:
+        self.last_sync_token is not None
+
     async def sync_tasks(self, response):
         if self.index:
             await self.index.commit_events()
@@ -540,7 +544,6 @@ class PanClient(AsyncClient):
         timeout = 30000
         sync_filter = {"room": {"state": {"lazy_load_members": True}}}
         next_batch = self.pan_store.load_token(self.server_name, self.user_id)
-        self.last_sync_token = next_batch
 
         # We don't store any room state so initial sync needs to be with the
         # full_state parameter. Subsequent ones are normal.
@@ -905,7 +908,7 @@ class PanClient(AsyncClient):
 
         self.handle_to_device_from_sync_body(body)
 
-        for room_id, room_dict in body["rooms"]["join"].items():
+        for room_id, room_dict in body.get("rooms", {}).get("join", {}).items():
             try:
                 if not self.rooms[room_id].encrypted:
                     logger.info(
@@ -920,7 +923,7 @@ class PanClient(AsyncClient):
                 # pan sync stream did. Let's assume that the room is encrypted.
                 pass
 
-            for event in room_dict["timeline"]["events"]:
+            for event in room_dict.get("timeline", {}).get("events", []):
                 if "type" not in event:
                     continue
 
