@@ -85,6 +85,7 @@ CORS_HEADERS = {
 
 class NotDecryptedAvailableError(Exception):
     """Exception that signals that no decrypted upload is available"""
+
     pass
 
 
@@ -120,7 +121,7 @@ class ProxyDaemon:
         self.hostname = self.homeserver.hostname
         self.store = PanStore(self.data_dir)
         accounts = self.store.load_users(self.name)
-        self.media_info = self.store.load_media(self.name)
+        self.media_info = self.store.load_media_cache(self.name)
         self.upload_info = self.store.load_upload(self.name)
 
         for user_id, device_id in accounts:
@@ -417,7 +418,9 @@ class ProxyDaemon:
         access_token = request.query.get("access_token", "")
 
         if not access_token:
-            access_token = request.headers.get("Authorization", "").strip("Bearer ")
+            access_token = request.headers.get("Authorization", "").replace(
+                "Bearer ", "", 1
+            )
 
         return access_token
 
@@ -459,6 +462,7 @@ class ProxyDaemon:
         data=None,  # type: bytes
         session=None,  # type: aiohttp.ClientSession
         token=None,  # type: str
+        use_raw_path=True,  # type: bool
     ):
         # type: (...) -> aiohttp.ClientResponse
         """Forward the given request to our configured homeserver.
@@ -473,6 +477,10 @@ class ProxyDaemon:
                 should be used to forward the request.
             token (str, optional): The access token that should be used for the
                 request.
+            use_raw_path (str, optional): Should the raw path be used from the
+            request or should we use the path and re-encode it. Some may need
+            their filters to be sanitized, this requires the parsed version of
+            the path, otherise we leave the path as is.
         """
         if not session:
             if not self.default_session:
@@ -481,9 +489,7 @@ class ProxyDaemon:
 
         assert session
 
-        path = urllib.parse.quote(
-            request.path
-        )  # re-encode path stuff like room aliases
+        path = request.raw_path if use_raw_path else urllib.parse.quote(request.path)
         method = request.method
 
         headers = CIMultiDict(request.headers)
@@ -762,7 +768,7 @@ class ProxyDaemon:
 
         try:
             response = await self.forward_request(
-                request, params=query, token=client.access_token
+                request, params=query, token=client.access_token, use_raw_path=False
             )
         except ClientConnectionError as e:
             return web.Response(status=500, text=str(e))
@@ -810,7 +816,9 @@ class ProxyDaemon:
             query["filter"] = request_filter
 
         try:
-            response = await self.forward_request(request, params=query)
+            response = await self.forward_request(
+                request, params=query, use_raw_path=False
+            )
         except ClientConnectionError as e:
             return web.Response(status=500, text=str(e))
 
@@ -834,9 +842,7 @@ class ProxyDaemon:
             body=await response.read(),
         )
 
-    def _get_upload_and_media_info(self, content_key, content):
-        content_uri = content[content_key]
-
+    def _get_upload_and_media_info(self, content_uri: str):
         try:
             upload_info = self.upload_info[content_uri]
         except KeyError:
@@ -846,7 +852,6 @@ class ProxyDaemon:
 
         self.upload_info[content_uri] = upload_info
 
-        content_uri = content[content_key]
         mxc = urlparse(content_uri)
         mxc_server = mxc.netloc.strip("/")
         mxc_path = mxc.path.strip("/")
@@ -859,13 +864,14 @@ class ProxyDaemon:
 
         return upload_info, media_info
 
-    async def _map_decrypted_uri(self, content_key, content, request, client):
-        upload_info, media_info = self._get_upload_and_media_info(content_key, content)
+    async def _decrypt_uri(self, content_uri, client):
+        upload_info, media_info = self._get_upload_and_media_info(content_uri)
         if not upload_info or not media_info:
             raise NotDecryptedAvailableError
 
-        response, decrypted_file = await self._load_decrypted_file(media_info.mxc_server, media_info.mxc_path,
-                                                                   upload_info.filename)
+        response, decrypted_file = await self._load_decrypted_file(
+            media_info.mxc_server, media_info.mxc_path, upload_info.filename
+        )
 
         if response is None and decrypted_file is None:
             raise NotDecryptedAvailableError
@@ -875,7 +881,7 @@ class ProxyDaemon:
 
         decrypted_upload, _ = await client.upload(
             data_provider=BufferedReader(BytesIO(decrypted_file)),
-            content_type=response.content_type,
+            content_type=upload_info.mimetype,
             filename=upload_info.filename,
             encrypt=False,
             filesize=len(decrypted_file),
@@ -884,9 +890,7 @@ class ProxyDaemon:
         if not isinstance(decrypted_upload, UploadResponse):
             raise NotDecryptedAvailableError
 
-        content[content_key] = decrypted_upload.content_uri
-
-        return content
+        return decrypted_upload.content_uri
 
     async def send_message(self, request):
         access_token = self.get_access_token(request)
@@ -900,12 +904,36 @@ class ProxyDaemon:
 
         room_id = request.match_info["room_id"]
 
-        # The room is not in the joined rooms list, just forward it.
         try:
             room = client.rooms[room_id]
             encrypt = room.encrypted
         except KeyError:
-            return await self.forward_to_web(request, token=client.access_token)
+            # The room is not in the joined rooms list, either the pan client
+            # didn't manage to sync the state or we're not joined, in either
+            # case send an error response.
+            if client.has_been_synced:
+                return web.json_response(
+                    {
+                        "errcode": "M_FORBIDDEN",
+                        "error": "You do not have permission to send the event.",
+                    },
+                    headers=CORS_HEADERS,
+                    status=403,
+                )
+            else:
+                logger.error(
+                    "The internal Pantalaimon client did not manage "
+                    "to sync with the server."
+                )
+                return web.json_response(
+                    {
+                        "errcode": "M_UNKNOWN",
+                        "error": "The pantalaimon client did not manage to sync with "
+                        "the server",
+                    },
+                    headers=CORS_HEADERS,
+                    status=500,
+                )
 
         # Don't encrypt reactions for now - they are weird and clients
         # need to support them like this.
@@ -923,10 +951,23 @@ class ProxyDaemon:
         # The room isn't encrypted just forward the message.
         if not encrypt:
             content_msgtype = content.get("msgtype")
-            if content_msgtype in ["m.image", "m.video", "m.audio", "m.file"] or msgtype == "m.room.avatar":
+            if (
+                content_msgtype in ["m.image", "m.video", "m.audio", "m.file"]
+                or msgtype == "m.room.avatar"
+            ):
                 try:
-                    content = await self._map_decrypted_uri("url", content, request, client)
-                    return await self.forward_to_web(request, data=json.dumps(content), token=client.access_token)
+                    content["url"] = await self._decrypt_uri(content["url"], client)
+                    if (
+                        "info" in content
+                        and "thumbnail_url" in content["info"]
+                        and content["info"]["thumbnail_url"] is not None
+                    ):
+                        content["info"]["thumbnail_url"] = await self._decrypt_uri(
+                            content["info"]["thumbnail_url"], client
+                        )
+                    return await self.forward_to_web(
+                        request, data=json.dumps(content), token=client.access_token
+                    )
                 except ClientConnectionError as e:
                     return web.Response(status=500, text=str(e))
                 except (KeyError, NotDecryptedAvailableError):
@@ -939,8 +980,13 @@ class ProxyDaemon:
         async def _send(ignore_unverified=False):
             try:
                 content_msgtype = content.get("msgtype")
-                if content_msgtype in ["m.image", "m.video", "m.audio", "m.file"] or msgtype == "m.room.avatar":
-                    upload_info, media_info = self._get_upload_and_media_info("url", content)
+                if (
+                    content_msgtype in ["m.image", "m.video", "m.audio", "m.file"]
+                    or msgtype == "m.room.avatar"
+                ):
+                    upload_info, media_info = self._get_upload_and_media_info(
+                        content["url"]
+                    )
                     if not upload_info or not media_info:
                         response = await client.room_send(
                             room_id, msgtype, content, txnid, ignore_unverified
@@ -953,10 +999,21 @@ class ProxyDaemon:
                             body=await response.transport_response.read(),
                         )
 
-                    media_content = media_info.to_content(content, upload_info.mimetype)
+                    media_info.to_content(content, upload_info.mimetype)
+                    if content["info"].get("thumbnail_url", False):
+                        (
+                            thumb_upload_info,
+                            thumb_media_info,
+                        ) = self._get_upload_and_media_info(
+                            content["info"]["thumbnail_url"]
+                        )
+                        if thumb_upload_info and thumb_media_info:
+                            thumb_media_info.to_thumbnail(
+                                content, thumb_upload_info.mimetype
+                            )
 
                     response = await client.room_send(
-                        room_id, msgtype, media_content, txnid, ignore_unverified
+                        room_id, msgtype, content, txnid, ignore_unverified
                     )
                 else:
                     response = await client.room_send(
@@ -1158,14 +1215,22 @@ class ProxyDaemon:
                     body=await response.transport_response.read(),
                 )
 
-            self.store.save_upload(self.name, response.content_uri, file_name, content_type)
+            self.store.save_upload(
+                self.name, response.content_uri, file_name, content_type
+            )
 
             mxc = urlparse(response.content_uri)
             mxc_server = mxc.netloc.strip("/")
             mxc_path = mxc.path.strip("/")
 
             logger.info(f"Adding media info for {mxc_server}/{mxc_path} to the store")
-            media_info = MediaInfo(mxc_server, mxc_path, maybe_keys["key"], maybe_keys["iv"], maybe_keys["hashes"])
+            media_info = MediaInfo(
+                mxc_server,
+                mxc_path,
+                maybe_keys["key"],
+                maybe_keys["iv"],
+                maybe_keys["hashes"],
+            )
             self.store.save_media(self.name, media_info)
 
             return web.Response(
@@ -1239,8 +1304,12 @@ class ProxyDaemon:
             return self._not_json
 
         try:
-            content = await self._map_decrypted_uri("avatar_url", content, request, client)
-            return await self.forward_to_web(request, data=json.dumps(content), token=client.access_token)
+            content["avatar_url"] = await self._decrypt_uri(
+                content["avatar_url"], client
+            )
+            return await self.forward_to_web(
+                request, data=json.dumps(content), token=client.access_token
+            )
         except ClientConnectionError as e:
             return web.Response(status=500, text=str(e))
         except (KeyError, NotDecryptedAvailableError):
@@ -1252,7 +1321,9 @@ class ProxyDaemon:
         file_name = request.match_info.get("file_name")
 
         try:
-            response, decrypted_file = await self._load_decrypted_file(server_name, media_id, file_name)
+            response, decrypted_file = await self._load_decrypted_file(
+                server_name, media_id, file_name
+            )
 
             if response is None and decrypted_file is None:
                 return await self.forward_to_web(request)
