@@ -23,7 +23,12 @@ from typing import List
 
 import attr
 import click
-from gi.repository import GLib
+
+try:
+    from gi.repository import GLib
+except ModuleNotFoundError:
+    from pgi.repository import GLib
+
 from prompt_toolkit import __version__ as ptk_version
 from prompt_toolkit import HTML, PromptSession, print_formatted_text
 from prompt_toolkit.completion import Completer, Completion, PathCompleter
@@ -690,9 +695,9 @@ class PanCtl:
         "the pantalaimon daemon."
     )
 )
-@click.version_option(version="0.9.1", prog_name="panctl")
+@click.version_option(version="0.10.4", prog_name="panctl")
 def main():
-    loop = asyncio.get_event_loop()
+    loop = asyncio.new_event_loop()
     glib_loop = GLib.MainLoop()
 
     try:
